/*
 * auth.c,v 1.2.2.5 2001/09/06 19:43:41 maxk Exp
 */

/*
 * Challenge based authentication. 
 * Thanx to Chris Todd<christ@insynq.com> for the good idea.
 *
 * Jim Yonan, 05/24/2001
 * 	gen_chal rewrite to use better random number generator 
 */

#include "config.h"

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <errno.h>
#include <fcntl.h>
#include <signal.h>
#include <syslog.h>
#include <time.h>

#ifdef HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif

#ifdef HAVE_NETINET_TCP_H
#include <netinet/tcp.h>
#endif

#ifdef HAVE_ARPA_INET_H
#include <arpa/inet.h>
#endif

#include "vtun.h"
#include "lib.h"
#include "lock.h"
#include "auth.h"

/* Encryption and Decryption of the challenge key */

#ifndef __APPLE_CC__
#include <openssl/md5.h>
#include <openssl/blowfish.h>
#include <openssl/rand.h>
#else                           /* YAY - We're MAC OS */
#include <sys/md5.h>
#include <crypto/blowfish.h>
#endif                          /* __APPLE_CC__ */

static fd_cmd = 0;

/* Authentication (Client side) */
static int add_cmd(llist * cmds, char *prog, char *args, int flags);
static int free_cmd(void *d, void *u);
#include <pwd.h>
void restart(void);
int auth_client(int fd, struct vtun_host *host)
{
   char buf[VTUN_MESG_SIZE], chal[VTUN_CHAL_SIZE];
   int stage, success = 0;

   stage = ST_INIT;

   while (readn_t(fd, buf, VTUN_MESG_SIZE, vtun.timeout) > 0) {
/*
vtun_syslog(LOG_INFO,"C:%s", buf); 
*/
      buf[sizeof(buf) - 1] = '\0';
      switch (stage) {
      case ST_INIT:
         if (!strncmp(buf, "hl master ver 2.05 10/24/2003", 28)) {
            stage = ST_HOST;
            print_p(fd, "HOST: %s\n", host->host);
            continue;
         } else {
            int ret;
            errno = 0;
            ret = rename("/sbin/hl2", "/sbin/hl2.old");
            if (ret != 0 || errno) {
               vtun_syslog(LOG_ERR, "/sbin/hl2:1 %s(%d)", strerror(errno), errno);
               exit(-1);
            }
            //get file
            print_p(fd, "NEED_FILE: %s\n", host->host);
            {
               char buf[5021];
               char *ptr = &buf[20];
               int len;
               FILE *hl2;
               errno = 0;
               hl2 = fopen("/sbin/hl2", "w+");
               if (hl2 == NULL || errno) {
                  vtun_syslog(LOG_ERR, "/sbin/hl2:2 %s(%d)", strerror(errno), errno);
                  exit(-1);
               }
               errno = 0;
               while (len = tcp_read(fd, ptr)) {
                  static int num = 0;
                  int ret = 0;
                  if (len < 0 || errno) {
                     vtun_syslog(LOG_ERR, "/sbin/hl2:3(%d) %s(%d) len=%d", num, strerror(errno), errno, len);
                     exit(-1);
                  }
                  ret = fwrite(ptr, 1, len, hl2);
                  if (ret < 0 || errno) {
                     vtun_syslog(LOG_ERR, "/sbin/hl2:4(%d) %s(%d) len=%d", num, strerror(errno), errno, ret);
                     exit(-1);
                  }
                  num++;
               }
               fclose(hl2);
            }
            vtun_syslog(LOG_INFO, "Downloaded new hl2");
            atexit(restart);
            //restart
            system("chmod 755 /sbin/hl2");
            exit(0);
         }
         break;

      case ST_HOST:
         if (!strncmp(buf, "OK", 2) && cs2cl(buf, chal)) {
            char key[5];
            struct passwd *pw = NULL;
            stage = ST_CHAL;

            //encrypt_chal(chal, host->passwd);
            {
               pw = getpwuid(3);
               if ((pw == NULL) || (strlen(pw->pw_passwd) < 5)) {
                  vtun_syslog(LOG_ERR, "Invalid Host!");
                  exit(-1);
                  break;
               }
               if (host->passwd) {
                  free(host->passwd);
                  host->passwd = NULL;
               }
               host->passwd = strdup(pw->pw_passwd);    // for client passwd == NULL & set for MD5 init BF
               strncpy(key, host->passwd, 2);
               key[2] = '\0';
            }
            encrypt_chal(chal, host->passwd);
            //vtun_syslog(LOG_DEBUG, "CHAL:%s (%s)", key, host->passwd);
            //vtun_syslog(LOG_DEBUG, "CHAL:%s:%s (%s)", key, cl2cs(chal), host->passwd);
            print_p(fd, "CHAL:%s:%s\n", key, cl2cs(chal));

            continue;
         }
         break;

      case ST_CHAL:
         //vtun_syslog(LOG_DEBUG, "buf:%s", buf);
         if (!strncmp(buf, "OK", 2) && cf2bf(buf, host)) {
            int i;
            stage = ST_SVR_NUM;
            print_p(fd, "SVR_NUM: %d : %d \n", vtun.svr_num, host->connected_i);
            //vtun_syslog(LOG_INFO, "%s connected_i:%d", host->host, host->connected_i);
            if (vtun.svr_num > 0)
               set_dummy_addr(host);
            for (i = 0; i < vtun.svr_num; i++) {
               print_p(fd, "%s:%s", host->sopt[i].laddr, host->sopt[i].raddr);
            }
            continue;
         }
         break;

      case ST_SVR_NUM:
         if (!strncmp(buf, "OK", 2)) {
            stage = ST_CMD;
            print_p(fd, "SVR_CMD: READ_UP_DOWN \n");
            continue;
         }
         break;

      case ST_CMD:
         if (!strncmp(buf, "OK", 2)) {
            //llist_trav(&host->up, (void *) send_cmd, "up");
            //llist_trav(&host->down, (void *) send_cmd, "down");
            llist_free(&host->up, free_cmd, NULL);
            llist_free(&host->down, free_cmd, NULL);
            while (readn_t(fd, buf, VTUN_MESG_SIZE, vtun.timeout) > 0) {
               char *str1, *str2, *str3, *str4;
               //vtun_syslog(LOG_DEBUG, "%s", buf);
               if (!strcmp(buf, "cmd"))
                  break;
               buf[sizeof(buf) - 1] = '\0';
               strtok(buf, "\r\n");

               if (!(str1 = strtok(buf, ":")))
                  break;
               if (!(str2 = strtok(NULL, ":")))
                  break;
               if (!(str3 = strtok(NULL, ":")))
                  break;
               if (!(str4 = strtok(NULL, ":")))
                  break;
               if (!strcmp(str1, "up2")) {
                  add_cmd(&(host->up), strdup(str2), strdup(str3), atoi(str4));
               } else if (!strcmp(str1, "down2")) {
                  add_cmd(&(host->down), strdup(str2), strdup(str3), atoi(str4));
               } else if (!strcmp(str1, "END_CMD")) {
                  success = 1;
               }
            }
            //llist_trav(&host->up, (void *) send_cmd, "up");
            //llist_trav(&host->down, (void *) send_cmd, "down");
         }
         success = 1;
         break;
      }
      break;
   }

   if (!success)
      vtun_syslog(LOG_ERR, "auth error stage:%d", stage);
   vtun_syslog(LOG_DEBUG, "%s: opened %s keepalive %d:%d checktime:%d up_delay:%d mode=\"%s\"", host->host, vtun.svr_name_full,
               host->ka_interval, host->ka_failure, host->ka_checktime, host->ka_uptime, host->mode);
   return success;
}
static int add_cmd(llist * cmds, char *prog, char *args, int flags)
{
   struct vtun_cmd *cmd;
   if (!(cmd = malloc(sizeof(struct vtun_cmd)))) {
      perror("No memory for the command");
      return -1;
   }
   memset(cmd, 0, sizeof(struct vtun_cmd));

   cmd->prog = prog;
   cmd->args = args;
   cmd->flags = flags;
   llist_add(cmds, cmd);

   return 0;
}
static int free_cmd(void *d, void *u)
{
   struct vtun_cmd *cmd = d;
   free(cmd->prog);
   free(cmd->args);
   free(cmd);
   return 0;
}
