/*
 * client.c,v 1.5.2.7 2002/04/25 09:19:50 bergolth Exp
 */

#include "config.h"
#include "vtun_socks.h"

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <fcntl.h>
#include <syslog.h>

#ifdef HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif

#ifdef HAVE_ARPA_INET_H
#include <arpa/inet.h>
#endif

#include "vtun.h"
#include "lib.h"
#include "llist.h"
#include "auth.h"
#include "compat.h"
#include "netlib.h"

static volatile sig_atomic_t client_term;
static void sig_term(int sig)
{
   vtun_syslog(LOG_INFO, "Terminated");
   client_term = VTUN_SIG_TERM;
}

extern svr_num_save;
void client(struct vtun_host *host)
{
   struct sockaddr_in my_addr, svr_addr;
   struct sigaction sa;
   int i, s, opt, reconnect;

   //vtun_syslog(LOG_INFO, "hl R:%s opened", VTUN_VER);

   memset(&sa, 0, sizeof(sa));
   sa.sa_handler = SIG_IGN;
   sa.sa_flags = SA_NOCLDWAIT;
   sigaction(SIGHUP, &sa, NULL);
   sigaction(SIGQUIT, &sa, NULL);
   sigaction(SIGPIPE, &sa, NULL);
   sigaction(SIGCHLD, &sa, NULL);
   sigaction(SIGALRM, &sa, NULL);


   sa.sa_handler = sig_term;
   sigaction(SIGTERM, &sa, NULL);
   sigaction(SIGINT, &sa, NULL);

   client_term = 0;
   reconnect = 0;
   while ((!client_term) || (client_term == VTUN_SIG_HUP)) {
      if (reconnect && (client_term != VTUN_SIG_HUP)) {
         if (vtun.persist || host->persist) {
            /* Persist mode. Sleep and reconnect. */
            //sleep(1);
            usleep(100);
         } else {
            /* Exit */
            break;
         }
      } else {
         reconnect = 1;
      }

      set_title("%s init initializing", host->host);

      /* Set server address */
      /*
         if( server_addr(&svr_addr, host, vtun.svr_name[0]) < 0 ) 
         continue;
       */




      /* 
       * Clear speed,pps and flags which will be supplied by server. 
       */
      vtun.svr_num = svr_num_save;
      host->spd_in = host->spd_out = 0;
      host->pps_in = host->pps_out = 0;
      host->flags &= VTUN_CLNT_MASK;

      io_init();

      set_title("%s connecting to %s", host->host, vtun.svr_name_full);
      //vtun_syslog(LOG_INFO, "Connecting to %s", vtun.svr_name_full);

      /* Set server address */
      for (i = 0; i < vtun.svr_num; i++) {
         /* Set local address */
         if (local_addr(&my_addr, host, 0) < 0) /* without option my_addr == 0.0.0.0 */
            continue;
         /* We have to create socket again every time
          * we want to connect, since STREAM sockets 
          * can be successfully connected only once.
          */
         if ((s = socket(AF_INET, SOCK_STREAM, 0)) == -1) {
            vtun_syslog(LOG_ERR, "Can't create socket. %s(%d)", strerror(errno), errno);
            continue;
         }
         /* Required when client is forced to bind to specific port */
         opt = 1;
         setsockopt(s, SOL_SOCKET, SO_REUSEADDR, &opt, sizeof(opt));
         if (bind(s, (struct sockaddr *) &my_addr, sizeof(my_addr))) {
            vtun_syslog(LOG_ERR, "Can't bind socket. %s(%d)", strerror(errno), errno);
            continue;
         }

         if (server_addr(&svr_addr, host, vtun.svr_name[i]) < 0)
            continue;
         if (connect_t(s, (struct sockaddr *) &svr_addr, 1) == 0) {
            host->connected_i = i + 1;
            break;
         }
         close(s);
      }
      {                         // connection try routine
         static int all_try = 0;
         if (i == vtun.svr_num) {       /* all try failed */
            if (all_try > 10)
               sleep(2);
            if (all_try > 100)
               sleep(4);
            if (all_try % 5 == 0)
               vtun_syslog(LOG_INFO, "%s: all try failed %s : %s(%d)", host->host, vtun.svr_name_full, strerror(errno), errno);
            all_try++;
         } else {
            if (auth_client(s, host)) {
               all_try = 0;
               //vtun_syslog(LOG_INFO, "%s: %s opened", host->host, vtun.svr_name_full);

               host->tcp_fd = s;

               /* Start the tunnel */
               client_term = tunnel(host);

               vtun_syslog(LOG_INFO, "%s: %s closed: term=%d : %s(%d)", host->host, vtun.svr_name_full, client_term,
                           strerror(errno), errno);
            } else {
               static int try = 0;
               try++;
               vtun_syslog(LOG_INFO, "%s: auth failed %s : try %d sleep %d", host->host, vtun.svr_name_full, try, try * 2);
               sleep(try * 2);
               if (try > 5)
                  exit(-1);
            }
         }
      }
      close(s);
      free_sopt(host->sopt);
   }

   vtun_syslog(LOG_INFO, "Exit");
   return;
}
