/*
 * linkfd.c,v 1.4.2.11 2002/04/25 09:19:50 bergolth Exp
 * For UDP MAT version (Full aggregation)
 */

#include "config.h"

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <signal.h>
#include <strings.h>
#include <string.h>
#include <errno.h>
#include <sys/time.h>
#include <syslog.h>
#include <time.h>

#ifdef HAVE_SYS_RESOURCE_H
#include <sys/resource.h>
#endif

#ifdef HAVE_SCHED_H
#include <sched.h>
#endif

#ifdef HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif

#include "vtun.h"
#include "linkfd.h"
#include "lib.h"
#include "driver.h"

char *get_local_addr(struct vtun_host *host, int i);
void Idle(int i, char *log);
void Send_Close_All(int i);
/* Host we are working with. 
 * Used by signal handlers that's why it is global. 
 */
struct vtun_host *lfd_host;

struct lfd_mod *lfd_mod_head = NULL, *lfd_mod_tail = NULL;

/* Modules functions*/

/* Add module to the end of modules list */
void lfd_add_mod(struct lfd_mod *mod)
{
   if (!lfd_mod_head) {
      lfd_mod_head = lfd_mod_tail = mod;
      mod->next = mod->prev = NULL;
   } else {
      lfd_mod_tail->next = mod;
      mod->prev = lfd_mod_tail;
      mod->next = NULL;
      lfd_mod_tail = mod;
   }
}

/*  Initialize and allocate each module */
int lfd_alloc_mod(struct vtun_host *host)
{
   struct lfd_mod *mod = lfd_mod_head;

   while (mod) {
      if (mod->alloc && (mod->alloc) (host))
         return 1;
      mod = mod->next;
   }

   return 0;
}

/* Free all modules */
int lfd_free_mod(void)
{
   struct lfd_mod *mod = lfd_mod_head;

   while (mod) {
      if (mod->free && (mod->free) ())
         return 1;
      mod = mod->next;
   }
   lfd_mod_head = lfd_mod_tail = NULL;
   return 0;
}

 /* Run modules down (from head to tail) */
inline int lfd_run_out(int len, char *in, char **out)
{
   register struct lfd_mod *mod;

   *out = in;
   for (mod = lfd_mod_head; mod && len > 0; mod = mod->next)
      if (mod->encode) {
         len = (mod->encode) (len, in, out);
         in = *out;
      }
   return len;
}

/* Run modules up (from tail to head) */
inline int lfd_run_in(int len, char *in, char **out)
{
   register struct lfd_mod *mod;

   *out = in;
   for (mod = lfd_mod_tail; mod && len > 0; mod = mod->prev)
      if (mod->decode) {
         len = (mod->decode) (len, in, out);
         in = *out;
      }
   return len;
}

/* Check if modules are accepting the data(down) */
inline int lfd_check_out(void)
{
   register struct lfd_mod *mod;
   int err = 1;

   for (mod = lfd_mod_head; mod && err > 0; mod = mod->next)
      if (mod->avail_encode)
         err = (mod->avail_encode) ();
   return err;
}

/* Check if modules are accepting the data(up) */
inline int lfd_check_in(void)
{
   register struct lfd_mod *mod;
   int err = 1;

   for (mod = lfd_mod_tail; mod && err > 0; mod = mod->prev)
      if (mod->avail_decode)
         err = (mod->avail_decode) ();

   return err;
}

/********** Linker *************/
/* Termination flag */
static volatile sig_atomic_t linker_term;

static void sig_term(int sig)
{
   vtun_syslog(LOG_INFO, "%s: closing by signal term", lfd_host->host);
   io_cancel();
   linker_term = VTUN_SIG_TERM;
}

static void sig_hup(int sig)
{
   vtun_syslog(LOG_INFO, "%s: retry by signal hup", lfd_host->host);
   io_cancel();
   linker_term = VTUN_SIG_HUP;
}

/* Statistic dump */
static unsigned long inline tv2ms(struct timeval tv)
{
   register unsigned long ms = (tv.tv_sec * 1000) + (tv.tv_usec / 1000);
   return ms ? ms : 1;
}

int live_host_num;
char live_host_str[10];
char live_remote_host[200];
unsigned long In, Out;
int In_pps, Out_pps;
struct timeval tm_IO;

int jitter[5 * 2 + 1];
char *bps(char *buf, unsigned long In, int gap);
void Set_live_remote_host()
{
   int i;
   char *p = live_remote_host;
   for (i = 0; i < vtun.svr_num; i++) {
      sprintf(p, "%s", lfd_host->sopt[i].raddr);
      if (i < vtun.svr_num - 1)
         strcat(p, " ");
      p += strlen(p);
   }
}

void Set_title()
{
   struct timeval tm;
   struct timeval tv;
   unsigned long gap;
   time_t gap2;
   gettimeofday(&tm, NULL);
   timersub(&tm, &tm_IO, &tv);
   gap = tv2ms(tv);
   if (gap < 5000)
      return;                   /* low gap */
   if (gap < 1)
      gap = 1;
   gap2 = time(NULL) - lfd_host->stat.tm_old;
   if (gap2 < 1)
      gap2 = 1;
   /*
      set_title
      ("%-6s %-5s %-4s %4luk %4luk (%4luk %4luk in %lu:%02lu) (%lu/%lu)(%d/%d)",
      lfd_host->host, lfd_host->sopt->dev, live_host_str,
      In * 8 / gap / 1000, Out * 8 / gap / 1000,
      lfd_host->stat.byte_in * 8 / gap2 / 1000,
      lfd_host->stat.byte_out * 8 / gap2 / 1000, gap2 / 60, gap2 % 60,
      lfd_host->stat.packet_loss, lfd_host->stat.packet_count, live_host_num, vtun.svr_num);
    */
   /*
      set_title
      ("%6s Download:%8lubps Upload:%8lubps (%d/%d) %5s %4s",
      lfd_host->host, 
      In * 8 / gap , Out * 8 / gap,
      live_host_num, vtun.svr_num
      lfd_host->sopt->dev, live_host_str,
      );
    */
   if (!vtun.is_client) {
      char buf[20], buf2[20];
      set_title
          ("%-7s %-5s %-4s %9s %9s  %4.1fpps %4.1fpps (%s) %s",
           lfd_host->host, lfd_host->sopt->dev, live_host_str,
           bps(buf, In, gap), bps(buf2, Out, gap), In_pps * 1000. / gap, Out_pps * 1000. / gap, live_remote_host, lfd_host->mode);
      /*
         set_title
         ("%-7s %-5s %-4s %9s %9s ",
         lfd_host->host, lfd_host->sopt->dev, live_host_str,
         bps(buf, In, gap), bps(buf2, Out, gap)
         );
       */
   }

   /*reset counter */
   In = Out = 0;
   In_pps = Out_pps = 0;
   tm_IO.tv_sec = tm.tv_sec;
   tm_IO.tv_usec = tm.tv_usec;
}

void log_stat(char *stat_in)
{
   static time_t tm;
   static char stm[22];
   time_t gap;
   int i;

   tm = time(NULL);
   gap = tm - lfd_host->stat.tm_old;
   if (gap < 10)
      return;                   /* low gap then return */

   strftime(stm, sizeof(stm) - 1, "%Y/%m/%d %H:%M:%S", localtime(&tm));
   /* 20040401 11:11:11 UUUU gap in_loss out_loss in_count out_count in_speed out_speed */
   fprintf(lfd_host->stat.file, "%s %s %3lu %4lu %5lu %3lu %6lu %s\n",
           stm, live_host_str, gap,
           lfd_host->stat.byte_in * 8 / gap / 1000,
           lfd_host->stat.byte_out * 8 / gap / 1000,
           lfd_host->stat.packet_loss, lfd_host->stat.packet_count, (stat_in ? stat_in : "0    0  END 0")
       );

   /* reset to zero */
   lfd_host->stat.byte_in = lfd_host->stat.byte_out = 0;
   lfd_host->stat.comp_in = lfd_host->stat.comp_out = 0;
   lfd_host->stat.packet_loss = lfd_host->stat.packet_count = 0;
   lfd_host->stat.tm_old = tm;
   for (i = 1; i <= vtun.svr_num; i++) {
      jitter[i] = 0;
   }
}
int idle[5 * 2 + 1];
int splite[5 * 2 + 1];
int fd[5 * 2 + 1], order[5 + 1];
char *buf[5 * 2 + 1], *out;

void send_log()
{
   static time_t tm;
   time_t gap;
   int i;

   tm = time(NULL);
   gap = tm - lfd_host->stat.tm_old;
   if (gap < 10)
      return;                   /* low gap then return */

   sprintf(buf[0], "%3lu %6lu %s %d", lfd_host->stat.packet_loss, lfd_host->stat.packet_count, live_host_str, gap);

   /*
      vtun_syslog(LOG_INFO,"from_server stat_req");
    */
   for (i = 1; i <= vtun.svr_num; i++) {
      jitter[i] = 0;
      if (idle[i] != -1) {
         if (udp_write2(fd[i], buf[0], VTUN_STAT_REP + strlen(buf[0]) + 1, 0) < 0)
            Idle(i, "stat_rep");
      }
   }

   /* reset to zero */
   lfd_host->stat.byte_in = lfd_host->stat.byte_out = 0;
   lfd_host->stat.comp_in = lfd_host->stat.comp_out = 0;
   lfd_host->stat.packet_loss = lfd_host->stat.packet_count = 0;
   lfd_host->stat.tm_old = tm;
}

void sig_alarm(int sig)
{
   int k;
   /* send request */
   for (k = 1; k <= vtun.svr_num; k++) {
      if (idle[k] != -1) {
         /*
            vtun_syslog(LOG_INFO,"STAT_REQ %d",k);
          */
         if (udp_write2(fd[k], buf[k], VTUN_STAT_REQ, 0) < 0)
            Idle(k, "send stat req error");
      }
   }
   alarm(VTUN_STAT_IVAL);
}

static void sig_usr1(int sig)
{
   /* Start Test module using sky00.cmd */
   /*
      #define VTUN_DIR "/hl"
      FILE *fp;
      char file[50];
      int do_num=0,k;
      sprintf(file,"%s/%.20s.do", VTUN_DIR, lfd_host->host);
      fp= fopen(file,"r");
      if( fp == NULL){
      vtun_syslog(LOG_INFO,"TEST_REQ can't read do file %s",file);
      return;
      }else{
      do_num= atoi(fgets(file,40 ,fp));
      if( do_num <1 || do_num > vtun.svr_num){
      vtun_syslog(LOG_INFO,"TEST_REQ Bad do line num %s",file);
      return;
      }
      vtun_syslog(LOG_INFO,"TEST_REQ Try line num %d", do_num);
      for( k=1; k<=vtun.svr_num; k++){
      if( idle[k] != -1){
      vtun_syslog(LOG_INFO,"TEST_REQ to line %d for test line %d",k, do_num);
      if( udp_write2(fd[k], buf[k], VTUN_TEST_REQ, do_num) < 0 ) Idle(k,"send test req error");
      }
      }
      }
    */
}


/* Link remote and local file descriptors */
int linkfd(struct vtun_host *host)
{
   struct sigaction sa, sa_oldterm, sa_oldint, sa_oldhup;
   int old_prio;

   lfd_host = host;

   old_prio = getpriority(PRIO_PROCESS, 0);
   setpriority(PRIO_PROCESS, 0, LINKFD_PRIO);

   /* Build modules stack */
   if (host->flags & VTUN_ZLIB)
      lfd_add_mod(&lfd_zlib);

   if (host->flags & VTUN_LZO)
      lfd_add_mod(&lfd_lzo);

   if (host->flags & VTUN_ENCRYPT)
      lfd_add_mod(&lfd_encrypt);

   if (host->flags & VTUN_SHAPE)
      lfd_add_mod(&lfd_shaper);

   /* pps shaper 15k default */
   if (host->flags & VTUN_PPS)
      lfd_add_mod(&lfd_pps);


   if (lfd_alloc_mod(host))
      return 0;

   memset(&sa, 0, sizeof(sa));
   sa.sa_handler = sig_term;
   sigaction(SIGTERM, &sa, &sa_oldterm);
   sigaction(SIGINT, &sa, &sa_oldint);
   sa.sa_handler = sig_hup;
   sigaction(SIGHUP, &sa, &sa_oldhup);

   /* It will use for test module */
   sa.sa_handler = sig_usr1;
   sigaction(SIGUSR1, &sa, NULL);
   /* Initialize statstic dumps when it is server */
   if ((host->flags & VTUN_STAT) && !vtun.is_client) {
      char file[40];

      sa.sa_handler = sig_alarm;
      sigaction(SIGALRM, &sa, NULL);

      sprintf(file, "%s/%.20s", VTUN_STAT_DIR, host->host);
      if ((host->stat.file = fopen(file, "a"))) {
         setvbuf(host->stat.file, NULL, _IOLBF, 0);
         alarm(VTUN_STAT_IVAL);
      } else
         vtun_syslog(LOG_ERR, "Can't open stats file %s", file);
   }

   io_init();

   lfd_linker_udp();

   if ((host->flags & VTUN_STAT) && !vtun.is_client) {
      alarm(0);
      if (host->stat.file)
         fclose(host->stat.file);
   }

   lfd_free_mod();

   sigaction(SIGTERM, &sa_oldterm, NULL);
   sigaction(SIGINT, &sa_oldint, NULL);
   sigaction(SIGHUP, &sa_oldhup, NULL);

   setpriority(PRIO_PROCESS, 0, old_prio);

   return linker_term;
}


int send_close_try_count;
int send_close[5 * 2 + 1];
void Send_Close_All(int i)
{
   int j;
   for (j = 1; j <= vtun.svr_num; j++) {
      if (idle[j] != -1) {
         /*
            vtun_syslog(LOG_INFO, "%s send close line %d using %d", lfd_host->host, i, j);
          */
         /* try close other side */
         udp_write2(fd[j], buf[j], VTUN_CONN_CLOSE, i);
         errno=0;
      }
   }
}

time_t fd_idle;
time_t fd_idle_last;
void Idle(int i, char *log)
{
   if (log)
      vtun_syslog(LOG_INFO, "%s: link down:%d (R:%s,L:%s) by:%s ",
                  lfd_host->host, i, lfd_host->sopt[i - 1].raddr, lfd_host->sopt[i - 1].laddr, log);
   live_host_num--;
   live_host_str[i - 1] = 'D';
   Set_live_remote_host();
   Set_title();
   idle[i] = -1;
   fd_idle = time(NULL);        /* time value for retry to make line up */
   fd_idle_last = time(NULL);   /* time value for skiping retry to make line up */
   Send_Close_All(i);
}

int Gap(int t, int base)
{
   int gap;
   gap = t - base;
   if (gap > MAX_SEQ / 2)
      gap -= MAX_SEQ;
   if (gap < -MAX_SEQ / 2)
      gap += MAX_SEQ;
   return gap;
}

int try_lineup[5 * 2 + 1];
int try_count[5*2 +1];
struct timeval fd_last[5 + 1], w_last[5 + 1];
int rseq;
int i_max;
int len, fl, wlen;
fd_set fdset;
int rlen[5 * 2 + 1];
int maxfd, rbuf[5 * 2 + 1];
int wseq = 0;
int next = 0;
int rseq_last = 1;
int read_count = 0;
unsigned long check_gap = 1500;
time_t start_time = 0;
int (*mod_in) (void);
int (*mod_out) (void);
int mat_in(void);
int mat_out(void);
int mat2_in(void);
int mat2_out(void);
int qat_in(void);
int qat_out(void);
int qmat_in(void);
int qmat_out(void);
int qat_num = 1;
int qat_size = 1000;
struct sockaddr_in saddr; // for client mode saddr

Set_Vars()                      // using ka_interval(*1000), ka_failure, ka_uptime(re-up time), ka_checktime(retry time)
{
   int i, j;

   rseq = 0;
   rseq_last = 1;
   next = 0;
   wseq = 0;
   check_gap = lfd_host->ka_interval;
   qat_num = vtun.svr_num;
   qat_size = 1000;

   i_max = vtun.svr_num + 1;
   send_close_try_count = 0;
   fd_idle = time(NULL);
   fd_idle_last = time(NULL) - 200;     // set for first up
   start_time = time(NULL) + 100;       // set for display up after start_time
   In = Out = 0;
   In_pps = Out_pps = 0;
   gettimeofday(&tm_IO, NULL);
   srandom(time(NULL));
   fd[0] = lfd_host->loc_fd;
   for (i = 1; i < i_max; i++) {
      fd[i] = lfd_host->rmt_fd[i - 1];
      gettimeofday(&fd_last[i], NULL);
      send_close[i] = 0;
   }
   for (i = 0; i < i_max; i++) {
      idle[i] = -1;             /*default to down */
      splite[i] = 0;
      rbuf[i] = 0;
      jitter[i] = 0;
      try_lineup[i] = 0;
      try_count[i] = 0;
      if (!(buf[i] = lfd_alloc(VTUN_FRAME_SIZE + VTUN_FRAME_OVERHEAD))) {
         vtun_syslog(LOG_ERR, "allocate error 1254");
         return 0;
      }
   }
   idle[0] = 0;
   idle[lfd_host->connected_i] = 0;
   live_host_num = 1;

   for (i = 0; i < i_max - 1; i++) {
      if (idle[i + 1] == 0)
         live_host_str[i] = 'U';
      else
         live_host_str[i] = 'D';
   }
   live_host_str[i] = '\0';
   Set_live_remote_host();
   // Set in out functions 
   if (!strncmp(lfd_host->mode, "MAT2", 4)) {
      mod_in = mat2_in;
      mod_out = mat2_out;
   } else if (!strncmp(lfd_host->mode, "MAT", 3)) {
      mod_in = mat_in;
      mod_out = mat_out;
   } else if (!strncmp(lfd_host->mode, "QAT", 3)) {
      mod_in = qat_in;
      mod_out = qat_out;
      {                         // for qat_num
         char buf[80], *ptr;
         strcpy(buf, lfd_host->mode);
         ptr = strstr(buf, ":");
         if (ptr) {
            qat_num = atoi(ptr + 1);
            if (qat_num < 1)
               qat_num = vtun.svr_num;
         } else {
            qat_num = vtun.svr_num;
         }
      }
      vtun_syslog(LOG_INFO, "%s: mode:%s (%d)", lfd_host->host, lfd_host->mode, qat_num);
   } else if (!strncmp(lfd_host->mode, "QMAT", 4)) {
      vtun_syslog(LOG_INFO, "%s: mode:%s", lfd_host->host, lfd_host->mode);
      mod_in = qmat_in;
      mod_out = qmat_out;
      {                         // for qmat_num
         char buf[80], *ptr;
         strcpy(buf, lfd_host->mode);
         ptr = strstr(buf, ":");
         if (ptr) {
            qat_size = atoi(ptr + 1);
            if (qat_size < 1)
               qat_size = 500;
         } else {
            qat_size = 500;
         }
         qat_num = 2;
      }
   } else {
      vtun_syslog(LOG_ERR, "%s: Invalid mode:%s", lfd_host->host, lfd_host->mode);
   }

}


Keep_Alive()
{
   int i, j;
   struct timeval fd_cur;
   struct timeval tv;

   if (lfd_host->flags & VTUN_KEEP_ALIVE) {
      /* First of All, send ECHO_REP when line is free by check_gap */
      for (i = 1; i < i_max; i++) {
         if (idle[i] != -1) {
            gettimeofday(&fd_cur, NULL);
            timersub(&fd_cur, &w_last[i], &tv);
            if (tv2ms(tv) > check_gap) {
               gettimeofday(&w_last[i], NULL);
               if (udp_write2(fd[i], buf[i], VTUN_ECHO_REP, 0) < 0)
                  Idle(i, "echo write error");
            }
         }
      }


      /* We are idle, lets check connection (changed for just read check) */
      for (i = 1; i < i_max; i++) {
         if (idle[i] != -1 && rbuf[i] == 0 && !FD_ISSET(fd[i], &fdset)) {       /* why use rbuf ?? */
            gettimeofday(&fd_cur, NULL);
            timersub(&fd_cur, &fd_last[i], &tv);        /* read gap */
            if (tv2ms(tv) > check_gap) {
               gettimeofday(&fd_last[i], NULL);

               if (++idle[i] > lfd_host->ka_failure) {
                  Idle(i, "network timeout");
                  /*
                     vtun_syslog(LOG_INFO,"%s Local network timeout line %d, Sending Close All",lfd_host->host, i);
                     idle[i]=0;
                     if(send_close[i] == 0){
                     send_close_try_count++;
                     send_close[i] = 1;
                     } 
                     Send_Close_All(i);
                     if( live_host_num <= send_close_try_count) live_host_num = 0;
                     break;
                   */

               }
            }
         }
      }
   }

}

Do_fl(int fl, int i)
{

   if (fl == VTUN_ECHO_REP) {
      /* Just ignore ECHO reply */
      if (idle[i] == -1) {
         if ((time(NULL) - fd_idle_last) > lfd_host->ka_uptime) {       /* del up lack */
            live_host_num++;
            live_host_str[i - 1] = 'U';
            Set_live_remote_host();
            if (time(NULL) > start_time)
               vtun_syslog(LOG_INFO,
                           "%s: link up:%d (R:%s,L:%s)", lfd_host->host, i, lfd_host->sopt[i - 1].raddr,
                           lfd_host->sopt[i - 1].laddr);
            Set_title();
            udp_write2(fd[i], buf[i], VTUN_ECHO_REP, 0);
            try_lineup[i] = 0;  /* reset to use again */
            try_count[i] = 0;  /* reset to use again */
            /*
               send_close[i]=0;
             */
            idle[i] = 0;
         } else {
            /*
               vtun_syslog(LOG_INFO,
               "%s network up SKIP p:%d (%s:%s)",
               lfd_host->host, i, lfd_host->sopt[i - 1].laddr, lfd_host->sopt[i - 1].raddr);
             */
         }
      } else {
         gettimeofday(&fd_last[i], NULL);
         idle[i] = 0;           /* idle 1,2 to 0 with ECHO_REP */
      }
   } else if (fl == VTUN_ECHO_REQ) {
      /* Send ECHO reply */
      if (udp_write2(fd[i], buf[i], VTUN_ECHO_REP, 0) < 0)
         Idle(i, "echo_req");
   } else if (fl == VTUN_STAT_REQ) {
      /* check qos */
      /*
         static last_live_host_num = 0;
         if (last_live_host_num != live_host_num) {
         run_cmd_sh("/hl/etc/link_num_changed", live_host_num);
         last_live_host_num = live_host_num;
         }
       */
      /* Send STAT reply */
      send_log();
   } else if (fl == VTUN_STAT_REP) {
      /*
         vtun_syslog(LOG_INFO,"from_client stat_rep %s", buf[i] );
       */
      log_stat(buf[i]);
   }
   /* Set rseq[0]=0 with ECHO SEQ reply for read */
   else if (fl == VTUN_TO_CLIENT) {
      int i_old = i;
      int i = rseq;
      int port;
      if (!try_lineup[i]) {
         /*
            vtun_syslog(LOG_INFO, "from_server SKIP port %d to make %d (using %d)", port, i, i_old);
          */
         return 0;
      }
      port = atoi(buf[i_old]);
      /*
         vtun_syslog(LOG_INFO, "from_server port %d to make %d (using %d)", port, i, i_old);
       */
      try_lineup[i] = 0;

      /* try connect */
      {
         int s = lfd_host->rmt_fd[i - 1];
         
         //using global saddr
         saddr.sin_addr.s_addr = inet_addr(lfd_host->sopt[i - 1].raddr);
         saddr.sin_port = lfd_host->sopt[i - 1].rport = port;

         if (connect(s, (struct sockaddr *) &saddr, sizeof(saddr))) {
               vtun_syslog(LOG_ERR, "Can't connect socket %s:%d", lfd_host->sopt[i - 1].raddr, lfd_host->sopt[i - 1].rport);
            /*
             */
            vtun_syslog(LOG_ERR, "connect() error 2312: %s (%d)", strerror(errno), errno);
            return -1;
         }
         if (udp_write2(fd[i], buf[i], VTUN_ECHO_REQ, 0)
             < 0) {
               vtun_syslog(LOG_ERR, "Check idle fd[%d] error after connect", i);
         }
      }
   } else if (fl == VTUN_TO_SERVER) {
      int i_old = i;
      int i = rseq;
      char *str1;
      int port;
      struct sockaddr_in saddr;
      str1 = strtok(buf[i_old], ":");
      port = atoi(&str1[strlen(str1) + 1]);
      /* if old==new ip then continue; to accept multi try */
      if (strcmp(lfd_host->sopt[i - 1].raddr, str1) == 0) {
         /*
            vtun_syslog(LOG_INFO, "from_client SKIP %s:%d to make %d (using %d)", str1, port, i, i_old);
          */
         return 0;
      }
      /*
         vtun_syslog(LOG_INFO, "from_client %s:%d to make %d (using %d)", str1, port, i, i_old);
       */

      free(lfd_host->sopt[i - 1].raddr);
      lfd_host->sopt[i - 1].raddr = strdup(str1);
      lfd_host->sopt[i - 1].rport = port;
      {                         /* make s and req to_client */
         {
            /* first server step  (make new socket and port) */
            unsigned port;
            int s, opt;
            if ((s = socket(AF_INET, SOCK_DGRAM, 0)) == -1) {
               vtun_syslog(LOG_ERR, "socket() error 3123");
               return -1;
            }

            opt = 1;
            setsockopt(s, SOL_SOCKET, SO_REUSEADDR, &opt, sizeof(opt));
            saddr.sin_addr.s_addr = inet_addr(lfd_host->sopt[i - 1].laddr);
            saddr.sin_port = 0;
            if (bind(s, (struct sockaddr *) &saddr, sizeof(saddr))) {
               vtun_syslog(LOG_ERR, "bind() error 4123");
               return -1;
            }

            opt = sizeof(saddr);
            if (getsockname(s, (struct sockaddr *) &saddr, &opt)) {
               vtun_syslog(LOG_ERR, "getsockname() error 5123");
               return -1;
            }
            port = saddr.sin_port;
            close(lfd_host->rmt_fd[i - 1]);     /* close old socket */
            lfd_host->rmt_fd[i - 1] = fd[i] = s;
            lfd_host->sopt[i - 1].lport = port;

            /* send request to client */
            {
               int k;
               for (k = 1; k < i_max; k++) {
                  if (idle[k] != -1) {
                     sprintf(buf[k], "%d", port);
                     /*
                        vtun_syslog(LOG_INFO, "to_client port %s %d (using %d)", buf[k], i, k);
                      */
                     if (udp_write2(fd[k], buf[k], VTUN_TO_CLIENT + strlen(buf[k]) + 1, i) < 0)
                        Idle(k, "send request write error");
                     break;
                  }
               }
            }
            /* try connect */
            saddr.sin_addr.s_addr = inet_addr(lfd_host->sopt[i - 1].raddr);
            saddr.sin_port = lfd_host->sopt[i - 1].rport;

            if (connect(s, (struct sockaddr *) &saddr, sizeof(saddr))) {
               /*
                  vtun_syslog(LOG_ERR, "Can't connect socket %s:%d", lfd_host->sopt[i - 1].raddr, lfd_host->sopt[i - 1].rport);
                */
               vtun_syslog(LOG_ERR, "connect() error 7123");
               return -1;
            }

         }                      /* End of Try routine */
      }
   } else if (fl == VTUN_BAD_FRAME) {
      vtun_syslog(LOG_ERR, "%s bad packet", lfd_host->host);
   } else if (fl == VTUN_CONN_CLOSE) {
      char j = rseq;
      /*
         vtun_syslog(LOG_ERR, "%s Received close line %d using %d", lfd_host->host, j, i);
       */
      if (idle[j] != -1) {
         /*
            if(send_close_try_count > 0) {
            send_close_try_count--;
            send_close[j]=0;
            }
          */
         Idle(j, "close by other side");
      }
   } else {
      vtun_syslog(LOG_ERR, "%s bad header fl:%X", lfd_host->host, fl);
      return 2;
   }
   return 1;
}

Check_Idle_Link()
{
   int i, j;
   /* check idle link if server mode */
   /* send request to client */
   //if( live_host_num == vtun.svr_num) return;
   if (!vtun.is_client && ((time(NULL) - fd_idle) > lfd_host->ka_checktime)) {
      for (i = 1; i < i_max; i++) {
         if (idle[i] == -1 && try_lineup[i]) {
            int k;
            for (k = 1; k < i_max; k++) {
               if (idle[k] != -1) {
                  sprintf(buf[k], "%d", lfd_host->sopt[k - 1].lport);
                  /*
                     vtun_syslog(LOG_INFO, "to_client port %s %d (using %d)", buf[k], i, k);
                   */
                  if (udp_write2(fd[k], buf[k], VTUN_TO_CLIENT + strlen(buf[k]) + 1, i) < 0)
                     Idle(k, "send request write error");
                  break;
               }
            }
         }
      }
   }

   /* check idle link if client mode */
   if (vtun.is_client && ((time(NULL) - fd_idle) > lfd_host->ka_checktime)) {
      static int j = 0;
      static int last_did =1; // for do one by one  with ka_checktime;
      j++;
      if (j > 3000) j = 1;

      while( idle[last_did] != -1){ // find next -1 line
         if( last_did > vtun.svr_num) {
               last_did=1;
               break;
         }
         last_did++;
      }

      //.vtun_syslog(LOG_DEBUG, "last_did=%d (2)", last_did);
      for (i = last_did; i < i_max; i++) { // need hack it work only one
         if (idle[i] == -1) {
            char *str;
            {
               str = get_local_addr(lfd_host, i - 1);
               if (strcmp(lfd_host->sopt[i - 1].laddr, str) == 0) {     /*try with same ip */
                  if (try_lineup[i]) {  /* retry for REQ_TO_SERVER : it will reset when line is up */
                     int k;
                     unsigned port = lfd_host->sopt[i - 1].lport;
                     for (k = 1; k < i_max; k++) {
                        if (idle[k] != -1) {
                           sprintf(buf[k], "%s:%d", str, port);
                           /*
                              vtun_syslog(LOG_ERR, "Retry to_server %s %d (use %d)", buf[k], port, i, k);
                            */
                           if (udp_write2(fd[k], buf[k], VTUN_TO_SERVER + strlen(buf[k]) + 1, i) < 0)
                              Idle(k, "send request write error");
                           break;
                        }
                     }
                  } else {
                     if (try_count[i] > 2) { // try 2 times and start restart_wan
                        run_cmd_sh("/etc/restart_wan", i);
                        vtun_syslog(LOG_DEBUG, "running /etc/restart_wan %d", i);
                        try_count[i] = 1;
                     }else{
                        vtun_syslog(LOG_DEBUG, "Check idle line %d j=%d", i, j);
                        if (udp_write2(fd[i], buf[i], VTUN_ECHO_REQ, 0) < 0) {
                           vtun_syslog(LOG_DEBUG, "Check idle fd[%d] error", i);
                        }
                        try_count[i] ++;
                     }
                  }
               } else {
                  try_lineup[i] = 1;
                     vtun_syslog(LOG_INFO, "Try New:%s Old:%s", str, lfd_host->sopt[i - 1].laddr);
                  /*
                   */
                  free(lfd_host->sopt[i - 1].laddr);
                  lfd_host->sopt[i - 1].laddr = str;
                  {
                     /* first step  (make new socket and port) */
                     unsigned port;
                     int s, opt;
                     if ((s = socket(AF_INET, SOCK_DGRAM, 0)) == -1) {
                        vtun_syslog(LOG_ERR, "socket() error 8124");
                        return -1;
                     }

                     opt = 1;
                     setsockopt(s, SOL_SOCKET, SO_REUSEADDR, &opt, sizeof(opt));
                     saddr.sin_addr.s_addr = inet_addr(lfd_host->sopt[i - 1].laddr);
                     saddr.sin_port = 0;
                     if (bind(s, (struct sockaddr *) &saddr, sizeof(saddr))) {
                        vtun_syslog(LOG_ERR, "bind() error 9123");
                        return -1;
                     }

                     opt = sizeof(saddr);
                     if (getsockname(s, (struct sockaddr *) &saddr, &opt)) {
                        vtun_syslog(LOG_ERR, "getsockname() error 1231");
                        return -1;
                     }
                     port = saddr.sin_port;
                     close(lfd_host->rmt_fd[i - 1]);    /* close old socket */
                     lfd_host->rmt_fd[i - 1] = fd[i] = s;
                     lfd_host->sopt[i - 1].lport = port;

                     /* send request to server */
                     {
                        int k;
                        for (k = 1; k < i_max; k++) {
                           if (idle[k] != -1) {
                              sprintf(buf[k], "%s:%d", str, port);
                              /*
                                 vtun_syslog(LOG_ERR, "to_server %s %d (use %d)", buf[k], port, i, k);
                               */
                              if (udp_write2(fd[k], buf[k], VTUN_TO_SERVER + strlen(buf[k]) + 1, i) < 0)
                                 Idle(k, "send request write error");
                              break;
                           }
                        }
                     }
                  }             /* End of Try routine */


               }
            }                   /* end strcmp */

         last_did++;
         break; // do just one and set next
         }
      }
      fd_idle = time(NULL);
   }
}


int lfd_linker_udp(void)
{
   int i, j;
   struct timeval wait_interval;
   struct timeval wait_interval_save;
   wait_interval_save.tv_sec = lfd_host->ka_interval / 1000;
   wait_interval_save.tv_usec = (lfd_host->ka_interval % 1000) * 1000;

   Set_Vars();

   /* send ECHO_REQ and set w_last value */
   for (i = 1; i < i_max; i++) {
      udp_write2(fd[i], buf[i], VTUN_ECHO_REQ, 0);
      gettimeofday(&w_last[i], NULL);
   }

   linker_term = 0;
   while (!linker_term) {
      errno = 0;
      /* Wait for data */
      FD_ZERO(&fdset);
      FD_SET(fd[0], &fdset);
      maxfd = fd[0];
      for (i = 1; i < i_max; i++) {
         if (rbuf[i] == 0)
            FD_SET(fd[i], &fdset);
         maxfd = (fd[i] > maxfd ? fd[i] : maxfd);
      }
      maxfd++;

      /* hm.. test for wait_interval changing.. bug?
         if( wait_interval.tv_sec != lfd_host->ka_interval / 1000 || wait_interval.tv_usec != lfd_host->ka_interval % 1000 * 1000){
         vtun_syslog(LOG_INFO, "miss sec:%d usec:%d", wait_interval.tv_sec, wait_interval.tv_usec);
         }
       */
      wait_interval.tv_sec = wait_interval_save.tv_sec;
      wait_interval.tv_usec = wait_interval_save.tv_usec;
      if ((len = select(maxfd, &fdset, NULL, NULL, &wait_interval)) < 0) {
         if (errno != EAGAIN && errno != EINTR)
            break;
         else
            continue;
      }

      Keep_Alive();
      Check_Idle_Link();        /*Try to UP */
      {
         int ret;
         ret = mod_in();
         if (ret != 0) {
            vtun_syslog(LOG_DEBUG, "mod_in: ret= %d", ret);
            break;
         }
         ret = mod_out();
         if (ret != 0) {
            vtun_syslog(LOG_DEBUG, "mod_out: ret= %d", ret);
            break;
         }
      }

      Set_title();
   }
   if (!vtun.is_client)
      log_stat(NULL);

   if (linker_term)
      vtun_syslog(LOG_INFO, "term: %ld", linker_term);

   if (errno)
      vtun_syslog(LOG_INFO, "error: %s (%d)", strerror(errno), errno);

   if (linker_term == VTUN_SIG_TERM) {
      lfd_host->persist = 0;
   }

   /* Notify other end about our close */

   for (i = 1; i < i_max; i++) {
      Send_Close_All(i);
   }
   for (i = i_max - 1; i >= 0; i--) {
      lfd_free(buf[i]);
   }
   return 0;
}

get_next()
{
   do {
      if (++next > vtun.svr_num)
         next = 1;
   } while (idle[next] == -1);
}

next_seq()
{
   wseq++;
   if (wseq < 1 || wseq > MAX_SEQ)
      wseq = 1;
}

#include "linkfd_mat.c"
#include "linkfd_mat2.c"
#include "linkfd_qat.c"
#include "linkfd_qmat.c"
