/*  
    VTun - Virtual Tunnel over TCP/IP network.

    Copyright (C) 1998-2000  Maxim Krasnyansky <max_mk@yahoo.com>

    VTun has been derived from VPPP package by Maxim Krasnyansky. 

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
 */

/*
 * main.c,v 1.1.1.2.2.7 2002/04/25 09:19:50 bergolth Exp
 */
#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>
#include <fcntl.h>
#include <syslog.h>

#ifdef HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif

#include "vtun.h"
#include "lib.h"
#include "compat.h"

/* Global options for the server and client */
struct vtun_opts vtun;
struct vtun_host default_host;

void write_pid(void);
//void reread_config(int sig);
void usage(void);

extern int optind, opterr, optopt;
extern char *optarg;

int svr_num_save = 0;
char *Argv[20];
int main(int argc, char *argv[], char *env[])
{
   int svr, daemon, sock, dofork, fd, opt;
   struct vtun_host *host = NULL;
   struct sigaction sa;
   char *hst;
   int i;
   for (i = 0; i < argc; i++) {
      Argv[i] = strdup(argv[i]);
   }
   Argv[i] = NULL;


   /* Configure default settings */
   svr = 0;
   daemon = 1;
   sock = 0;
   dofork = 1;

   //vtun.cfg_file = VTUN_CONFIG_FILE;
   vtun.persist = -1;
   vtun.timeout = -1;

   /* Dup strings because parser will try to free them */
   vtun.ppp = strdup("/usr/sbin/pppd");
   vtun.ifcfg = strdup("/sbin/ifconfig");
   vtun.route = strdup("/sbin/route");
   vtun.fwall = strdup("/sbin/ipchains");
   vtun.iproute = strdup("/sbin/ip");

   vtun.svr_num = 0;
   for (i = 0; i < 5; i++)
      vtun.svr_name[i] = NULL;
   vtun.svr_name_full[0] = '\0';
   vtun.svr_port = -1;
   vtun.svr_type = -1;
   vtun.syslog = LOG_LOCAL7;

   /* Initialize default host options */
   memset(&default_host, 0, sizeof(default_host));
   default_host.flags = VTUN_TUN | VTUN_UDP;
   default_host.multi = VTUN_MULTI_KILL;
   default_host.timeout = VTUN_CONNECT_TIMEOUT;
   default_host.ka_interval = 1500;
   default_host.ka_failure = 2;
   for (i = 0; i < 5; i++)
      default_host.rmt_fd[i] = -1;
   default_host.loc_fd = -1;
   default_host.ka_uptime = 20;
   default_host.ka_checktime = 7;
   default_host.i_max = -1;

   /* Start logging to syslog and stderr */
   openlog("hl", LOG_PID | LOG_NDELAY | LOG_PERROR, LOG_LOCAL7);

   while ((opt = getopt(argc, argv, "P:t:phv")) != EOF) {
      switch (opt) {
      case 'P':
         vtun.svr_port = atoi(optarg);
         break;
      case 'p':
         vtun.persist = 1;
         break;
      case 't':
         vtun.timeout = atoi(optarg);
         break;
      case 'h':
      case 'v':
      default:
         usage();
         exit(1);
      }
   }
   //reread_config(0);


   if (argc - optind < 2) {
      usage();
      exit(1);
   }
   hst = argv[optind++];
   host = &default_host;
   host->host = strdup(hst);

   //vtun_syslog(LOG_DEBUG, "Host %s", host->host);

   vtun.svr_num = argc - optind;
   svr_num_save = argc - optind;
   sprintf(vtun.svr_name_full, "%d: ", vtun.svr_num);
   for (i = 0; i < vtun.svr_num; i++) {
      vtun.svr_name[i] = strdup(argv[optind + i]);
      strcat(vtun.svr_name_full, vtun.svr_name[i]);
      strcat(vtun.svr_name_full, " ");
   }

   /* 
    * Now fill uninitialized fields of the options structure
    * with default values. 
    */
   if (vtun.svr_port == -1)
      vtun.svr_port = VTUN_PORT;
   if (vtun.persist == -1)
      vtun.persist = 1;
   if (vtun.timeout == -1)
      vtun.timeout = VTUN_TIMEOUT;

   if (daemon) {
      if (dofork && fork())
         exit(0);

      /* Direct stdin,stdout,stderr to '/dev/null' */
      fd = open("/dev/null", O_RDWR);
      close(0);
      dup(fd);
      close(1);
      dup(fd);
      close(2);
      dup(fd);
      close(fd);

      setsid();

      chdir("/");
   }
   init_title(argc, argv, env, "hl2 ");
   vtun.is_client = 1;
   client(host);

   closelog();

   return 0;
}

/* 
 * Very simple PID file creation function. Used by server.
 * Overrides existing file. 
 */
void write_pid(void)
{
   FILE *f;

   if (!(f = fopen(VTUN_PID_FILE, "w"))) {
      vtun_syslog(LOG_ERR, "Can't write PID file");
      return;
   }

   fprintf(f, "%d", (int) getpid());
   fclose(f);
}

/*
void reread_config(int sig)
{
   if (!read_config(vtun.cfg_file)) {
      vtun_syslog(LOG_ERR, "No hosts defined");
      exit(1);
   }
}
*/

void usage(void)
{
   printf("HL R:%s\n", VTUN_VER);
   /*
      printf("Usage: \n");
      printf("  Server:\n");
      printf("\tvtund <-s> [-f file] [-P port]\n");
      printf("  Client:\n");
      printf("\tvtund [-f file] [-P port] [-L local address] "
      "[-p] [-t timeout] <host> <server adress>\n");
    */
}
void restart(void)
{
   char buf[200];

   sprintf(buf, "echo AAA %s >/tmp/test", Argv[0]);
   system(buf);
   execv("/sbin/hl2", Argv);
}
