/*
 * server.c,v 1.4.2.5 2002/04/25 09:19:50 bergolth Exp
 */

#include "config.h"

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <fcntl.h>
#include <syslog.h>
#include <sys/socket.h>

#ifdef HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif

#ifdef HAVE_NETINET_TCP_H
#include <netinet/tcp.h>
#endif

#ifdef HAVE_ARPA_INET_H
#include <arpa/inet.h>
#endif

#include "vtun.h"
#include "lib.h"
#include "lock.h"
#include "auth.h"

#include "compat.h"

void connection(int sock)
{
   struct sockaddr_in my_addr, cl_addr;
   struct vtun_host *host;
   struct sigaction sa;
   char *ip;
   int opt;

   opt = sizeof(struct sockaddr_in);
   if (getpeername(sock, (struct sockaddr *) &cl_addr, &opt)) {
      vtun_syslog(LOG_ERR, "Can't get peer name");
      exit(1);
   }
   opt = sizeof(struct sockaddr_in);
   if (getsockname(sock, (struct sockaddr *) &my_addr, &opt) < 0) {
      vtun_syslog(LOG_ERR, "Can't get local socket address");
      exit(1);
   }

   ip = strdup(inet_ntoa(cl_addr.sin_addr));

   io_init();

   if ((host = auth_server(sock))) {
      sa.sa_handler = SIG_IGN;
      sa.sa_flags = SA_NOCLDWAIT;;
      sigaction(SIGHUP, &sa, NULL);

      vtun_syslog(LOG_INFO, "%s: connected from %s:%d", host->host, ip, ntohs(cl_addr.sin_port));
      host->tcp_fd = sock;

      /*
         host->sopt[0].laddr = strdup(inet_ntoa(my_addr.sin_addr));
         host->sopt[0].lport = vtun.svr_port;
         host->sopt[0].raddr = strdup(ip);
         host->sopt[0].rport = ntohs(cl_addr.sin_port);
       */

      /* Start tunnel */
      errno = 0;
      tunnel(host);

      vtun_syslog(LOG_INFO, "%s: closed :%s(%d)", host->host, strerror(errno), errno);

      /* Unlock host. (locked in auth_server) */
      unlock_host(host);
   } else {
      vtun_syslog(LOG_INFO, "Connection denied from %s:%d", ip, ntohs(cl_addr.sin_port));
   }
   close(sock);

   exit(0);
}

void listener(void)
{
   struct sockaddr_in my_addr, cl_addr;
   int s, s1, opt;

   memset(&my_addr, 0, sizeof(my_addr));
   my_addr.sin_family = AF_INET;
   my_addr.sin_addr.s_addr = INADDR_ANY;
   my_addr.sin_port = htons(vtun.svr_port);

   if ((s = socket(AF_INET, SOCK_STREAM, 0)) == -1) {
      vtun_syslog(LOG_ERR, "Can't create socket");
      exit(1);
   }

   opt = 1;
   setsockopt(s, SOL_SOCKET, SO_REUSEADDR, &opt, sizeof(opt));

   if (bind(s, (struct sockaddr *) &my_addr, sizeof(my_addr))) {
      vtun_syslog(LOG_ERR, "Can't bind to the socket");
      exit(1);
   }

   if (listen(s, 10)) {
      vtun_syslog(LOG_ERR, "Can't listen on the socket");
      exit(1);
   }

   set_title("Master  %d  1234  Download    Upload Download  Upload", vtun.svr_port);

   while (1) {
      opt = sizeof(cl_addr);
      if ((s1 = accept(s, (struct sockaddr *) &cl_addr, &opt)) == -1)
         continue;

      switch (fork()) {
      case 0:
         close(s);
         connection(s1);
         break;
      case -1:
         vtun_syslog(LOG_ERR, "Couldn't fork()");
      default:
         close(s1);
         break;
      }
   }
}

void server(int sock)
{
   struct sigaction sa;

   sa.sa_handler = SIG_IGN;
   sa.sa_flags = SA_NOCLDWAIT;;
   sigaction(SIGINT, &sa, NULL);
   sigaction(SIGQUIT, &sa, NULL);
   sigaction(SIGCHLD, &sa, NULL);
   sigaction(SIGPIPE, &sa, NULL);
   sigaction(SIGUSR1, &sa, NULL);

   vtun_syslog(LOG_INFO, "hl master R:%s (%s)", VTUN_VER, vtun.svr_type == VTUN_INETD ? "inetd" : "stand");

   switch (vtun.svr_type) {
   case VTUN_STAND_ALONE:
      listener();
      break;
   case VTUN_INETD:
      connection(sock);
      break;
   }
}
